// ----------------------------------------------
#include <assert.h>
// ----------------------------------------------
#include <GL/gl.h>
// ----------------------------------------------
#include "VertexStorage.h"
#include "CGPUmesh_GL_types.h"
#include "CGPUmesh_GL_funcs.h"
// ----------------------------------------------
namespace GPUmesh {
// ----------------------------------------------
template <class VertexFormat,class T_IndexType=Loki::NullType> 
class CompilePolicy_GL_Calls
{
public:
	
  typedef struct { unsigned char data[MVF_sizeof<VertexFormat>::value]; } vertex_data;
  typedef FormatDescriptor<VertexFormat>                                  vertex_format_desc;
  typedef VertexStorageInterlaced<VertexFormat,vertex_data>               vertex_storage;
  typedef std::vector<T_IndexType>                                        index_storage;

  typedef struct
  {
    int                             iPrimType;
    const vertex_storage           *storage;
    const index_storage            *indices;
  }mesh_descriptor;

  // ---------------------------------

  static void compile(int primtype,
                      vertex_storage& storage,
                      index_storage&  indices,
                      mesh_descriptor& _mesh)
  {
    _mesh.iPrimType=primtype;
    _mesh.storage=&storage;
    _mesh.indices=&indices;
  }

  static void     bind(const mesh_descriptor& m)
  {
  }

  static void     unbind(const mesh_descriptor& m)
  {
  }

  static void     free(const mesh_descriptor& m)
  {
  }

  static void     draw_vertex(const mesh_descriptor& m,int i)
  {
     assert(i >= 0 && i < (int)m.storage->m_Vertices.size());

    if (vertex_format_desc::has_normal)
    {
      (GLFuncs<typename vertex_format_desc::type_normal::type>::getNormalFct(vertex_format_desc::type_normal::components - 1))
        ((typename vertex_format_desc::type_normal::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_normal]));
    }
    if (vertex_format_desc::has_color0)
    {
      (GLFuncs<typename vertex_format_desc::type_color0::type>::getColor0Fct(vertex_format_desc::type_color0::components - 1))
        ((typename vertex_format_desc::type_color0::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_color0]));
    }
    if (vertex_format_desc::has_color1)
    {
      (GLFuncs<typename vertex_format_desc::type_color1::type>::getAttributeFct(vertex_format_desc::type_color1::components - 1))
        (4,(typename vertex_format_desc::type_color1::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_color1]));
    }

    if (vertex_format_desc::has_texcoord0)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord0::type>::getTexCoordnFct(vertex_format_desc::type_texcoord0::components - 1))
        (GL_TEXTURE0_ARB,(typename vertex_format_desc::type_texcoord0::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord0]));
    }

    if (vertex_format_desc::has_texcoord1)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord1::type>::getTexCoordnFct(vertex_format_desc::type_texcoord1::components - 1))
        (GL_TEXTURE1_ARB,(typename vertex_format_desc::type_texcoord1::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord1]));
    }

    if (vertex_format_desc::has_texcoord2)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord2::type>::getTexCoordnFct(vertex_format_desc::type_texcoord2::components - 1))
        (GL_TEXTURE2_ARB,(typename vertex_format_desc::type_texcoord2::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord2]));
    }

    if (vertex_format_desc::has_texcoord3)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord3::type>::getTexCoordnFct(vertex_format_desc::type_texcoord3::components - 1))
        (GL_TEXTURE3_ARB,(typename vertex_format_desc::type_texcoord3::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord3]));
    }

    if (vertex_format_desc::has_texcoord4)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord4::type>::getTexCoordnFct(vertex_format_desc::type_texcoord4::components - 1))
        (GL_TEXTURE4_ARB,(typename vertex_format_desc::type_texcoord4::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord4]));
    }

    if (vertex_format_desc::has_texcoord5)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord5::type>::getTexCoordnFct(vertex_format_desc::type_texcoord5::components - 1))
        (GL_TEXTURE5_ARB,(typename vertex_format_desc::type_texcoord5::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord5]));
    }

    if (vertex_format_desc::has_texcoord6)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord6::type>::getTexCoordnFct(vertex_format_desc::type_texcoord6::components - 1))
        (GL_TEXTURE6_ARB,(typename vertex_format_desc::type_texcoord6::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord6]));
    }

    if (vertex_format_desc::has_texcoord7)
    {
      (GLFuncs<typename vertex_format_desc::type_texcoord7::type>::getTexCoordnFct(vertex_format_desc::type_texcoord7::components - 1))
        (GL_TEXTURE7_ARB,(typename vertex_format_desc::type_texcoord7::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_texcoord7]));
    }

    if (vertex_format_desc::has_vertex)
    {
      (GLFuncs<typename vertex_format_desc::type_vertex::type>::getVertexFct(vertex_format_desc::type_vertex::components - 1))
        ((typename vertex_format_desc::type_vertex::type *)&(m.storage->m_Vertices[i].data[vertex_format_desc::offset_vertex]));
    }
  }

  static void     draw(const mesh_descriptor& m)
  {
    float test[]={1.0,0.0,0.0,1.0};

    glBegin(m.iPrimType);
    for (unsigned int i=0;i<m.storage->m_Vertices.size();i++)
    {
      draw_vertex(m,i);
    }
    glEnd();
  }
};

// ----------------------------------------------
// Indexed mesh
// ----------------------------------------------

template <class T_VertexFormat,class T_IndexType> 
class CompilePolicy_Indexed_GL_Calls : public CompilePolicy_GL_Calls<T_VertexFormat,T_IndexType>
{
public:
  typedef CompilePolicy_GL_Calls<T_VertexFormat,T_IndexType> parent_policy;
  typedef typename parent_policy::mesh_descriptor            mesh_descriptor;
public:
  static void     draw(const mesh_descriptor& m)
  {
    glBegin(m.iPrimType);
    for (unsigned int j=0;j<m.indices->size();j++)
    {
      draw_vertex(m,(*m.indices)[j]);
    }
    glEnd();
  }
};
// ----------------------------------------------
// define class with this compile policy
// ----------------------------------------------
template <class T_VertexFormat> 
class GPUmesh_GL_Calls
    : public GPUmesh_generic< CompilePolicy_GL_Calls , T_VertexFormat > {};

template <class T_VertexFormat,class T_IndexType> 
class GPUmesh_Indexed_GL_Calls
    : public GPUmesh_generic< CompilePolicy_Indexed_GL_Calls , T_VertexFormat , T_IndexType > {};
// ----------------------------------------------
} // namespace GPUmesh
// ----------------------------------------------
