// ----------------------------------------------
#define BUFFER_OFFSET(o) (((char *)NULL) + o)
// ----------------------------------------------
#ifndef GPUMESH_NO_GLUX
#  include <glux.h>
GLUX_REQUIRE(GL_ARB_vertex_buffer_object);
#endif
// ----------------------------------------------
#include "VertexStorage.h"
#include "CGPUmesh_GL_types.h"
#include "CGPUmesh_GL_funcs.h"
#include "CGPUmesh_primitives.h"
// ----------------------------------------------
namespace GPUmesh {
  // ----------------------------------------------
  template <class VertexFormat,class T_IndexType> 
  class CompilePolicy_GL_VBO
  {
  public:

    typedef struct { unsigned char data[MVF_sizeof<VertexFormat>::value]; } vertex_data;
    typedef FormatDescriptor<VertexFormat>                                  vertex_format_desc;
    typedef VertexStorageSeparateAttributes<VertexFormat,vertex_data>       vertex_storage;
    typedef std::vector<T_IndexType>                                        index_storage;

    enum {nb_vertex_attrib = Loki::TL::Length<VertexFormat>::value};

    class mesh_descriptor
    {
    public:
      GLuint uiVB;
      int    iVertices;
      int    iPrimType;
      int    offsets[nb_vertex_attrib];
    };

    // ---------------------------------

    static void compile(int primtype,
      vertex_storage& storage,
      index_storage&  indices,
      mesh_descriptor& _mesh)
    {
      const int       vfsz=Loki::TL::Length<VertexFormat>::value;

      int nbv=storage.m_iVertices;
      _mesh.iVertices=nbv;
      _mesh.iPrimType=primtype;
      // compute offsets
      for (int i=0;i<vfsz;i++)
      {
        int offset=vertex_format_desc::offset(i);
        _mesh.offsets[i]=offset*nbv;
      }
      // send to graphics card
      glGenBuffersARB(1,&_mesh.uiVB);
      glBindBufferARB(GL_ARRAY_BUFFER_ARB,_mesh.uiVB);
      glBufferDataARB(GL_ARRAY_BUFFER_ARB,nbv*vertex_format_desc::size_of,NULL,GL_STATIC_DRAW_ARB);
      for (int i=0;i<vertex_storage::nb_attributes;i++)
      {
        glBufferSubDataARB(GL_ARRAY_BUFFER_ARB,
          _mesh.offsets[i],
          storage.m_Attributes[i].size(),
          &(storage.m_Attributes[i].front()));
      }

      // free memory
      storage.clear();
      indices.clear();
    }

    // ---------------------------------

    static void     bind(const mesh_descriptor& m)
    {
      glBindBufferARB(GL_ARRAY_BUFFER_ARB,m.uiVB);

      if (vertex_format_desc::has_vertex)
      {
        glVertexPointer(vertex_format_desc::type_vertex::components,GLTypes<typename vertex_format_desc::type_vertex::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_vertex]));
        glEnableClientState(GL_VERTEX_ARRAY);
      }
      if (vertex_format_desc::has_normal)
      {
        glNormalPointer(GLTypes<typename vertex_format_desc::type_normal::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_normal]));
        glEnableClientState(GL_NORMAL_ARRAY);
      }
      if (vertex_format_desc::has_color0)
      {
        glColorPointer(vertex_format_desc::type_color0::components,GLTypes<typename vertex_format_desc::type_color0::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_color0]));
        glEnableClientState(GL_COLOR_ARRAY);
      }
      if (vertex_format_desc::has_color1)
      {
        glVertexAttribPointerARB(4,vertex_format_desc::type_color1::components,GLTypes<typename vertex_format_desc::type_color1::type>::gl_define,
          false,0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_color1]));
        glEnableVertexAttribArrayARB(4);
      }

      if (vertex_format_desc::has_texcoord0)
      {
        glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord0::components,GLTypes<typename vertex_format_desc::type_texcoord0::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord0]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord1)
      {
        glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord1::components,GLTypes<typename vertex_format_desc::type_texcoord1::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord1]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord2)
      {
        glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord2::components,GLTypes<typename vertex_format_desc::type_texcoord2::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord2]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord3)
      {
        glClientActiveTextureARB(GL_TEXTURE3_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord3::components,GLTypes<typename vertex_format_desc::type_texcoord3::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord3]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord4)
      {
        glClientActiveTextureARB(GL_TEXTURE4_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord4::components,GLTypes<typename vertex_format_desc::type_texcoord4::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord4]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord5)
      {
        glClientActiveTextureARB(GL_TEXTURE5_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord5::components,GLTypes<typename vertex_format_desc::type_texcoord5::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord5]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord6)
      {
        glClientActiveTextureARB(GL_TEXTURE6_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord6::components,GLTypes<typename vertex_format_desc::type_texcoord6::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord6]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord7)
      {
        glClientActiveTextureARB(GL_TEXTURE7_ARB);
        glTexCoordPointer(vertex_format_desc::type_texcoord7::components,GLTypes<typename vertex_format_desc::type_texcoord7::type>::gl_define,
          0,BUFFER_OFFSET(m.offsets[vertex_format_desc::idx_texcoord7]));
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      glClientActiveTextureARB(GL_TEXTURE0_ARB);
    }

    // ---------------------------------

    static void     unbind(const mesh_descriptor& m)
    {
      if (vertex_format_desc::has_vertex)
        glDisableClientState(GL_VERTEX_ARRAY);
      if (vertex_format_desc::has_normal)
        glDisableClientState(GL_NORMAL_ARRAY);
      if (vertex_format_desc::has_color0)
        glDisableClientState(GL_COLOR_ARRAY);
      if (vertex_format_desc::has_color1)
        glDisableVertexAttribArrayARB(4);
      if (vertex_format_desc::has_texcoord0)
      {
        glClientActiveTextureARB(GL_TEXTURE0_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
      if (vertex_format_desc::has_texcoord1)
      {
        glClientActiveTextureARB(GL_TEXTURE1_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
      if (vertex_format_desc::has_texcoord2)
      {
        glClientActiveTextureARB(GL_TEXTURE2_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
      if (vertex_format_desc::has_texcoord3)
      {
        glClientActiveTextureARB(GL_TEXTURE3_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      if (vertex_format_desc::has_texcoord4)
      {
        glClientActiveTextureARB(GL_TEXTURE4_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
      if (vertex_format_desc::has_texcoord5)
      {
        glClientActiveTextureARB(GL_TEXTURE5_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
      if (vertex_format_desc::has_texcoord6)
      {
        glClientActiveTextureARB(GL_TEXTURE6_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }
      if (vertex_format_desc::has_texcoord7)
      {
        glClientActiveTextureARB(GL_TEXTURE7_ARB);
        glDisableClientState(GL_TEXTURE_COORD_ARRAY);
      }

      glClientActiveTextureARB(GL_TEXTURE0_ARB);
    }

    // ---------------------------------

    static void     free(const mesh_descriptor& m)
    {
      glDeleteBuffersARB(1,&m.uiVB);
    }

    // ---------------------------------

    static void     draw(const mesh_descriptor& m)
    {
      glDrawArrays(m.iPrimType,0,m.iVertices);
    }
  };

  // ----------------------------------------------
  // Indexed mesh
  // ----------------------------------------------

  template <class T_VertexFormat,class T_IndexType> 
  class CompilePolicy_Indexed_GL_VBO : public CompilePolicy_GL_VBO<T_VertexFormat,T_IndexType>
  {
    // TODO: add static check with index type not working (e.g. int)

  public:

    typedef CompilePolicy_GL_VBO<T_VertexFormat,T_IndexType>                  parent_policy;
    typedef struct { unsigned char data[MVF_sizeof<T_VertexFormat>::value]; } vertex_data;
    typedef FormatDescriptor<T_VertexFormat>                                  vertex_format_desc;
    typedef VertexStorageSeparateAttributes<T_VertexFormat,vertex_data>       vertex_storage;
    typedef std::vector<T_IndexType>                                          index_storage;

    enum {nb_vertex_attrib = Loki::TL::Length<T_VertexFormat>::value};

    class mesh_descriptor : public parent_policy::mesh_descriptor
    {
    public:
      GLuint uiIB;
      int    iIndices;
    };

    // ---------------------------------

    static void compile(int primtype,
      vertex_storage& storage,
      index_storage&  indices,
      mesh_descriptor& _mesh)
    {
      const int       vfsz=Loki::TL::Length<T_VertexFormat>::value;

      // create vertex buffer
      int nbv=storage.m_iVertices;
      _mesh.iVertices=nbv;
      _mesh.iPrimType=primtype;
      // compute offsets
      for (int i=0;i<vfsz;i++)
      {
        int offset=vertex_format_desc::offset(i);
        _mesh.offsets[i]=offset*nbv;
      }
      // send to graphics card
      glGenBuffersARB(1,&_mesh.uiVB);
      glBindBufferARB(GL_ARRAY_BUFFER_ARB,_mesh.uiVB);
      glBufferDataARB(GL_ARRAY_BUFFER_ARB,nbv*vertex_format_desc::size_of,NULL,GL_STATIC_DRAW_ARB);
      for (int i=0;i<vertex_storage::nb_attributes;i++)
      {
        glBufferSubDataARB(GL_ARRAY_BUFFER_ARB,
          _mesh.offsets[i],
          storage.m_Attributes[i].size(),
          &(storage.m_Attributes[i].front()));
      }
      // create index buffer
      _mesh.iIndices=indices.size();
      glGenBuffersARB(1,&_mesh.uiIB);
      glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,_mesh.uiIB);
      glBufferDataARB(GL_ELEMENT_ARRAY_BUFFER_ARB,
        indices.size()*sizeof(GLTypes<T_IndexType>::gl_define),
        &(indices.front()),
        GL_STATIC_DRAW_ARB);

      // free memory
      storage.clear();
      indices.clear();
    }

    // ---------------------------------

    static void     bind(const mesh_descriptor& m)
    {
      parent_policy::bind(m);
      glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,m.uiIB);
    }

    // ---------------------------------

    static void     unbind(const mesh_descriptor& m)
    {
      parent_policy::unbind(m);
      glBindBufferARB(GL_ELEMENT_ARRAY_BUFFER_ARB,0);
    }

    // ---------------------------------

    static void     free(const mesh_descriptor& m)
    {
      parent_policy::free(m);
      glDeleteBuffersARB(1,&m.uiIB);
    }

    // ---------------------------------

    static void     draw(const mesh_descriptor& m)
    {
      glDrawElements(m.iPrimType,m.iIndices,GLTypes<T_IndexType>::gl_define,NULL);
    }

  };
  // ----------------------------------------------
  // define class with this compile policy
  // ----------------------------------------------
  template <class T_VertexFormat> 
  class GPUmesh_GL_VBO
    : public GPUmesh_generic< CompilePolicy_GL_VBO , T_VertexFormat > {};

  template <class T_VertexFormat,class T_IndexType> 
  class GPUmesh_Indexed_GL_VBO
    : public GPUmesh_generic< CompilePolicy_Indexed_GL_VBO , T_VertexFormat , T_IndexType > {};
  // ----------------------------------------------
} // namespace GPUmesh
// ----------------------------------------------
