// ----------------------------------------------
#ifndef __CRENDERMESH_D3D_SORTTYPELIST__
#define __CRENDERMESH_D3D_SORTTYPELIST__
// ----------------------------------------------
#include <TypeManip.h>
#include "VertexFormat.h"
// ----------------------------------------------
namespace GPUmesh {
// ----------------------------------------------
// Associates a value to mvf item for D3D
// ----------------------------------------------
template <class T> struct MVF_D3D_item_keys;
template <> struct MVF_D3D_item_keys<Loki::NullType> { enum {value = 999}; };
template <> struct MVF_D3D_item_keys<MVF_BASE_VERTEX>    { enum {value =   0}; };
template <> struct MVF_D3D_item_keys<MVF_BASE_NORMAL>    { enum {value =   1}; };
template <> struct MVF_D3D_item_keys<MVF_BASE_COLOR0>    { enum {value =   2}; };
template <> struct MVF_D3D_item_keys<MVF_BASE_COLOR1>    { enum {value =   3}; };
template <> struct MVF_D3D_item_keys<MVF_BASE_TEXCOORD0> { enum {value =   4}; };
template <> struct MVF_D3D_item_keys<MVF_BASE_TEXCOORD1> { enum {value =   5}; };
template <class T> struct MVF_D3D_item_keys          
{ 
private:
  enum {temp = MVF_D3D_item_keys<T::base_attribute>::value};
public:
  enum {value = temp};
};
// ----------------------------------------------
// Find min item
// ----------------------------------------------

template <class VertexFormat> class MVF_D3D_find_min;

template <> class MVF_D3D_find_min<Loki::NullType>
{
public:
  typedef Loki::NullType Result;
};

template <class VertexFormat> class MVF_D3D_find_min
{
private:

  ASSERT_TYPELIST(VertexFormat);

  typedef typename VertexFormat::Tail Tail;
  typedef typename VertexFormat::Head Head;

  typedef typename MVF_D3D_find_min<Tail>::Result next_min;

public:

  typedef typename Loki::Select<(MVF_D3D_item_keys<Head>::value < MVF_D3D_item_keys<next_min>::value),
                                Head,
                                next_min>::Result
                   Result;
};
// ----------------------------------------------
// Sorts vertex format to correspond to D3D FVF ordering
// ----------------------------------------------
template <class VertexFormat> class MVF_D3D_sort_items;

template <> class MVF_D3D_sort_items<Loki::NullType>
{
public:
  typedef Loki::NullType Result;
};

template <class VertexFormat> class MVF_D3D_sort_items
{
private:

  ASSERT_TYPELIST(VertexFormat);

  typedef typename VertexFormat::Head Head;
  typedef typename VertexFormat::Tail Tail;

  typedef MVF_D3D_find_min<VertexFormat>::Result        min;
  typedef Loki::TL::Replace<Tail,min,Head>::Result      next;

public:

  typedef Loki::Typelist< min , MVF_D3D_sort_items<next>::Result > Result;

};
// ----------------------------------------------
} // namespace GPUmesh
// ----------------------------------------------
#endif
// ----------------------------------------------
